//
// Example of reflective water via GLSL programs
//
// Author: Alex V. Boreskoff <alexboreskoff@mtu-net.ru>, <steps3d@narod.ru>
//

#include    "libExt.h"

#include    <glut.h>
#include    <stdio.h>
#include    <stdlib.h>

#include    "libTexture.h"
#include    "libTexture3D.h"
#include    "TypeDefs.h"
#include	"Vector4D.h"
#include    "Vector3D.h"
#include    "Vector2D.h"
#include    "boxes.h"
#include    "GlslProgram.h"
#include    "Camera.h"
#include	"FrameBuffer.h"
#include	"VertexBuffer.h"
#include	"utils.h"
#include	"fpTexture.h"

#define		SIZE		16
#define		TEX_SIZE	128

Vector3D    eye   ( 0, 0, 0 );  		// camera position
unsigned	noiseMap;
unsigned	reflectionMap;
float		yaw   = 0;
float		pitch = 0;
float		roll  = 0;
float		amplitude   = 0.1;
float		amplitude2  = 0.2;
bool        useFilter   = true;
bool		useRotation = true;
int		 	object      = 0;
Vector3D	eta ( 1.05, 1.1, 1.15 );

GLenum		mapFormat;				// texture format used for vertex map
unsigned	verticesMap;			// map with resulting vertices
unsigned	normalsMap;				// map with resulting normals

FrameBuffer	buffer  ( TEX_SIZE, TEX_SIZE );
GlslProgram	r2vb;
GlslProgram program;
Camera		camera ( eye, yaw, pitch, roll );	// camera to be used

VertexBuffer	* vertexBuffer;		// vertex coordinates
VertexBuffer    * normalBuffer;		// normal buffer
VertexBuffer	* indexBuffer;		// vertex index buffer

void reshape ( int w, int h );


float	* buf = (float *) malloc ( TEX_SIZE * TEX_SIZE * 4 * sizeof ( float ) );

void	createIndexBuffer ()
{
	int * data = (int *) malloc ( (TEX_SIZE-1)*(TEX_SIZE-1)*6*sizeof (int) );
	int	  k    = 0;
	
	for ( int i = 0; i < TEX_SIZE - 1; i++ )
		for ( int j = 0; j < TEX_SIZE - 1; j++ )
		{
			data [k]   = i + TEX_SIZE*j;				// first triangle (i,j)-(i+1,j) - (i,j+1)
			data [k+1] = i + TEX_SIZE*j + 1;
			data [k+2] = i + TEX_SIZE*j + TEX_SIZE;
			
			data [k+3] = i + TEX_SIZE*j + 1;			// second triangle (i+1,j)-(i+1,j+1) - (i,j+1)
			data [k+4] = i + TEX_SIZE*j + TEX_SIZE + 1;
			data [k+5] = i + TEX_SIZE*j + TEX_SIZE;
			
			k         += 6;
		}
		
	indexBuffer = new VertexBuffer ();
	
	indexBuffer -> bind    ( GL_ELEMENT_ARRAY_BUFFER_ARB );
	indexBuffer -> setData ( (TEX_SIZE - 1) * (TEX_SIZE - 1) * 6 * sizeof ( int ), data, GL_STATIC_DRAW );
	indexBuffer -> unbind  ();
	
	free ( data );
}

void	createVertexBuffer ()
{
	vertexBuffer = new VertexBuffer ();
	
	vertexBuffer -> bind    ( GL_PIXEL_PACK_BUFFER_ARB );
	vertexBuffer -> setData ( buffer.getWidth () * buffer.getHeight () * 4 * sizeof ( float ), NULL, GL_DYNAMIC_DRAW );
	vertexBuffer -> unbind  ();
}

void	createNormalBuffer ()
{
	normalBuffer = new VertexBuffer ();
	
	normalBuffer -> bind    ( GL_PIXEL_PACK_BUFFER_ARB );
	normalBuffer -> setData ( buffer.getWidth () * buffer.getHeight () * 4 * sizeof ( float ), NULL, GL_DYNAMIC_DRAW );
	normalBuffer -> unbind  ();
}

//
// create vertex coordinates using r2vb
//
void	buildVertices ()
{
	GLenum	buffers [] = { GL_COLOR_ATTACHMENT0_EXT, GL_COLOR_ATTACHMENT1_EXT };

														// load source image map
	glActiveTextureARB ( GL_TEXTURE0_ARB );
    glBindTexture      ( GL_TEXTURE_3D, noiseMap );

	r2vb.bind      ();									// perform rendering to FBO texture
	buffer.bind    ();
	glDrawBuffers  ( 2, buffers );
	startOrtho     ( TEX_SIZE, TEX_SIZE );
	drawQuad       ( TEX_SIZE, TEX_SIZE );
	endOrtho       ();
	r2vb.unbind    ();
	
														// copy vertex data from texture to vertex buffer
	glReadBuffer           ( GL_COLOR_ATTACHMENT0_EXT );
	vertexBuffer -> bind   ( GL_PIXEL_PACK_BUFFER_ARB );
	glReadPixels           ( 0, 0, TEX_SIZE, TEX_SIZE, GL_RGBA, GL_FLOAT, NULL );
	vertexBuffer -> unbind ();

	glReadBuffer           ( GL_COLOR_ATTACHMENT1_EXT );
	normalBuffer -> bind   ( GL_PIXEL_PACK_BUFFER_ARB );
	glReadPixels           ( 0, 0, TEX_SIZE, TEX_SIZE, GL_RGBA, GL_FLOAT, NULL );
	normalBuffer -> unbind ();

	glReadPixels ( 0, 0, TEX_SIZE, TEX_SIZE, GL_RGBA, GL_FLOAT, buf );
	
	buffer.unbind          ();
}

void displayBoxes ()
{
    glMatrixMode ( GL_MODELVIEW );
    glPushMatrix ();

    glActiveTextureARB ( GL_TEXTURE1_ARB );
    glBindTexture      ( GL_TEXTURE_3D, noiseMap );

    glActiveTextureARB ( GL_TEXTURE0_ARB );

	drawCubicMappedBox ( Vector3D ( SIZE, SIZE, SIZE ), reflectionMap );

	glPopMatrix  ();
    glPushMatrix ();

    glPushClientAttrib   ( GL_CLIENT_VERTEX_ARRAY_BIT );
	
	glEnableClientState  ( GL_VERTEX_ARRAY );
	vertexBuffer -> bind ( GL_ARRAY_BUFFER_ARB );
	glVertexPointer      ( 4, GL_FLOAT, 0, NULL );
	
	glEnableClientState  ( GL_TEXTURE_COORD_ARRAY );
	normalBuffer -> bind ( GL_ARRAY_BUFFER_ARB );
	glTexCoordPointer    ( 4, GL_FLOAT, 0, NULL );
	
	indexBuffer  -> bind ( GL_ELEMENT_ARRAY_BUFFER_ARB );
	
	program.bind   ();
	glDrawElements ( GL_TRIANGLES, (TEX_SIZE-1)*(TEX_SIZE-1)*6, GL_UNSIGNED_INT, 0 );
	program.unbind ();

	glPopClientAttrib ();

    glPopMatrix ();
}

void display ()
{
	buildVertices ();

    glClear ( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

	camera.apply ();

	glMatrixMode ( GL_TEXTURE );
	glPushMatrix ();
	glRotatef    ( 90, 1, 0, 0 );

	glMatrixMode ( GL_MODELVIEW );

    displayBoxes ();

	glMatrixMode ( GL_TEXTURE );
	glPopMatrix  ();

	glMatrixMode ( GL_MODELVIEW );

    glutSwapBuffers ();
}

void reshape ( int w, int h )
{
	camera.setViewSize ( w, h, 60 );
	camera.apply       ();
}

void key ( unsigned char key, int x, int y )
{
    if ( key == 27 || key == 'q' || key == 'Q' )        // quit requested
        exit ( 0 );

   	if ( key == 'w' || key == 'W' )
   		camera.moveBy ( camera.getViewDir () * 0.2 );
   	else
   	if ( key == 'x' || key == 'X' )
   		camera.moveBy ( -camera.getViewDir () * 0.2 );
   	else
   	if ( key == 'a' || key == 'A' )
   		camera.moveBy ( -camera.getSideDir () * 0.2 );
   	else
   	if ( key == 'd' || key == 'D' )
   		camera.moveBy ( camera.getSideDir () * 0.2 );
	else
	if ( key == '+' )
		amplitude += 0.02;
	else
	if ( key == '-' )
		amplitude -= 0.02;
	else
	if ( key == '*' )
		amplitude2 += 0.02;
	else
	if ( key == '/' )
		amplitude2 -= 0.02;
		
	glutPostRedisplay ();
}

void    specialKey ( int key, int x, int y )
{
    if ( key == GLUT_KEY_UP )
        yaw += M_PI / 90;
    else
    if ( key == GLUT_KEY_DOWN )
        yaw -= M_PI / 90;
	else
    if ( key == GLUT_KEY_RIGHT )
        roll += M_PI / 90;
    else
    if ( key == GLUT_KEY_LEFT )
        roll -= M_PI / 90;

	camera.setEulerAngles ( yaw, pitch, roll );

    glutPostRedisplay ();
}

void	mouseFunc ( int x, int y )
{
	static	int	lastX = -1;
	static	int	lastY = -1;

	if ( lastX == -1 )				// not initialized
	{
		lastX = x;
		lastY = y;
	}

	yaw  -= (y - lastY) * 0.02;
	roll += (x - lastX) * 0.02;

	lastX = x;
	lastY = y;

	camera.setEulerAngles ( yaw, pitch, roll );

	glutPostRedisplay ();
}

void    animate ()
{
	static	float	lastTime = 0.0;
	float			time     = 0.001f * glutGet ( GLUT_ELAPSED_TIME );

    lastTime = time;

    program.bind ();
    program.setUniformVector ( "eyePos",      Vector4D ( camera.getPos (), 1 ) );
    program.setUniformFloat  ( "time",        time );
    program.setUniformFloat  ( "amplitude",   amplitude );
    program.setUniformFloat  ( "amplitude2",  amplitude2 );
    program.unbind ();
	
	r2vb.bind            ();
    r2vb.setUniformFloat ( "time",        time );
    r2vb.setUniformFloat ( "amplitude",   amplitude );
	r2vb.unbind          ();

    glutPostRedisplay ();
}

int main ( int argc, char * argv [] )
{
                                // initialize glut
    glutInit            ( &argc, argv );
    glutInitDisplayMode ( GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH );
    glutInitWindowSize  ( 512, 512 );


                                // create window
    glutCreateWindow ( "OpenGL reflective water with R2VB and EMBM" );

                                // register handlers
    glutDisplayFunc       ( display    );
    glutReshapeFunc       ( reshape    );
    glutKeyboardFunc      ( key        );
    glutSpecialFunc       ( specialKey );
    glutIdleFunc          ( animate    );
    glutPassiveMotionFunc ( mouseFunc  );

    init           ();
    initExtensions ();

	assertExtensionsSupported ( "ARB_pixel_buffer_object" );

	if ( (mapFormat = fpFormatWithPrecision ( 32 ) ) == 0 )
	{
		printf ( "Floating-point textures not supported !\n" );
		
		return 1;
	}
	
	if ( !FrameBuffer :: isSupported () )
	{
		printf ( "Fbo not supported\n" );
		
		return 1;
	}
	
	if ( !GlslProgram :: isSupported () )
	{
		printf ( "GLSL not supported.\n" );

		return 1;
	}

	unsigned vertMap = buffer.createColorRectTexture ( GL_RGBA, mapFormat );
	unsigned normMap = buffer.createColorRectTexture ( GL_RGBA, mapFormat );
	
	buffer.create ();
	buffer.bind   ();
	if ( !buffer.attachColorTexture ( GL_TEXTURE_RECTANGLE_ARB, vertMap, 0 ) )
		printf ( "buffer error with color attachment 0\n");

	if ( !buffer.attachColorTexture ( GL_TEXTURE_RECTANGLE_ARB, normMap, 1 ) )
		printf ( "buffer error with color attachment 1\n");

	if ( !buffer.isOk () )
		printf ( "Error with framebuffer\n" );
					
	buffer.unbind ();
	
    if ( !program.loadShaders ( "water-3.vsh", "water-3.fsh" ) )
    {
        printf ( "Error loading shaders:\n%s\n", program.getLog ().c_str () );

        return 2;
    }

    if ( !r2vb.loadShaders ( "r2vb-1.vsh", "r2vb-1.fsh" ) )
    {
        printf ( "Error loading shaders:\n%s\n", r2vb.getLog ().c_str () );

        return 3;
    }

	const char * faces [6] =
	{
		"../../Textures/Cubemaps/skyrt.bmp",
		"../../Textures/Cubemaps/skylf.bmp",
		"../../Textures/Cubemaps/skydn.bmp",
		"../../Textures/Cubemaps/skyup.bmp",
		"../../Textures/Cubemaps/skybk.bmp",
		"../../Textures/Cubemaps/skyfr.bmp",
	};

	noiseMap      = createTexture3D ( false, "../../Textures/Noise/noise-3D.dds" );
	reflectionMap = createCubeMap   ( true, faces );

    r2vb.bind ();
	r2vb.setTexture ( "noiseMap", 0 );
    r2vb.unbind ();

    program.bind ();
    program.setTexture ( "envMap",   0 );
    program.setTexture ( "noiseMap", 1 );
    program.unbind ();

	camera.setRightHanded ( false );

	createVertexBuffer ();
	createNormalBuffer ();
	createIndexBuffer  ();
	
	printf ( "Use + and - to control waves vertex amplitude\nUse * and / to control bump amplitude\nUse WXAD to move camera\n" );

    glutMainLoop ();

    return 0;
}
